/*
 * SW_CUDA_CPU.cu
 *
 *  Created on: Nov 12, 2009
 *      Author: fritz
 */

#include "cuda_sw.cuh"

#include <cmath>

#include "gpu/cuda_sw_gpu.cuh"
#include "functions/cuda_functions.cuh"
#include "parser/reference_parser.h"
void Sw(Reference & scaffold, Reads & reads, Parameters parameters, bool rev,vector< vector< int > > BSlookup) {

	long threads = 256; //fixed threads per block (number chosen to be the optimum)


	//calcualte the memory needed per thread:
	int size = reads.get_max_length() + parameters.corridor_with;
	long mem_parameters = 9;
	long mem_per_thread = sizeof(float) * 2 * parameters.corridor_with + reads.get_max_length() * sizeof(char) + sizeof(float) + size * sizeof(char);
	//									matrix						     reads     					mem_result     scaff
	long mem2 = parameters.memsize - (mem_parameters * sizeof(float));


	//how many processors can we start?
	unsigned int blocks = (mem2 / (threads * mem_per_thread));
	blocks = blocks - parameters.mps;
	if (blocks > 2250) {
		blocks = 2250;
	}

	if (reads.get_max_length() > 200) {
		blocks = blocks / 2;
	}
	if (reads.get_max_length() > 700 && blocks > 300) {
		blocks = 300;
	}

	int procs = threads * blocks;

	long mem_reads = reads.get_max_length() * procs;
	long mem_scaff = size * procs;
	long mem_result = procs;
	long mem_matrix = 2 * procs * parameters.corridor_with;

	char * reads_cpu = new char[mem_reads];
	char * scaff_cpu = new char[mem_scaff];
	float * results_cpu = new float[mem_result];
	float * parameters_cpu = new float[mem_parameters];

	memset(results_cpu, -1.0f, mem_result * sizeof(float));

	//parameters needed on the GPU

	parameters_cpu[0] = (float) reads.get_max_length();
	parameters_cpu[1] = (float) size;
	parameters_cpu[2] = parameters.match[0];
	parameters_cpu[3] = parameters.mismatch[0];
	parameters_cpu[4] = parameters.gap_read[0];
	parameters_cpu[5] = (float) parameters.corridor_with; //corridor
	parameters_cpu[6] = parameters.gap_scaffold[0];
	parameters_cpu[7] = parameters.T_match[0];
	parameters_cpu[8] = parameters.T_mismatch[0];

//	CUDA_SAFE_CALL(cudaMemcpyToSymbol(param, parameters_cpu,mem_parameters* sizeof(float)));

	float * param;
	CUDA_SAFE_CALL(cudaMalloc((void**)&param,mem_parameters*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemcpy(param,parameters_cpu,mem_parameters*sizeof(float),cudaMemcpyHostToDevice));

	char * scaffold_gpu;
	char * reads_gpu;
	float * matrix_gpu;
	float * results_gpu;



	CUDA_SAFE_CALL(cudaMalloc((void**)&scaffold_gpu,mem_scaff*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&reads_gpu,mem_reads*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&results_gpu,mem_result*sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&matrix_gpu,mem_matrix*sizeof(float)));
	checkCUDAError("after mallocs");

	CUDA_SAFE_CALL(cudaMemset(matrix_gpu,0.0f,mem_matrix*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemset(results_gpu,-1.0f,mem_result*sizeof(float)));
	checkCUDAError("after memcpy");

	unsigned int shared_mem = (sizeof(char) + sizeof(unsigned int) + 2 * sizeof(float)) * threads;

	dim3 dimBlock(threads, 1);
	dim3 dimGrid(blocks, 1); //per matrix 1 thread!


	//needed for prefiltering:
	iterators_str iter;
	iter.readID = 0;
	iter.size = size;
	iter.seqlength = reads.get_max_length();
	iter.match = reads.matches.begin();
	iter.proc = procs;
	if (parameters.seedword_hits > 0) {
		iter.seed_tresh = parameters.seedword_hits;
	}
//	else {
//		iter.seed_tresh = call_parameter(reads.forward[0].size());
//	}

	iter.shift = round((double) (parameters.corridor_with) / 2.0);
	iter.tresh = round((double) (parameters.corridor_with / 3.0));
	iter.numofSeq = reads.forward.size();

	if (reads.get_direction()) {
		iter.read_pos = reads.forward.begin();
		iter.mapps = reads.maped_reads.begin();
	} else if (!reads.get_direction()) {
		iter.read_pos = reads.backward.begin();
		iter.mapps = reads.maped_reads_back.begin();
	}

	int t1 = 0;
	int t2 = 0;

	vector<HelpStruct> helper1;
	vector<HelpStruct> helper2;

	HelpStruct temp;
	temp.id = 0;
	temp.position = 0;
	temp.scaff_id = 0;

	for (size_t i = 0; i <= procs; i++) {
		helper1.push_back(temp);
		helper2.push_back(temp);
	}

	map<int, list<ids_str> > liste;
	list<ids_str> tmp;
	for (int i = scaffold.get_ID(); i < scaffold.name.size(); i++) {
		liste[i] = tmp;
	}

	long amount = 0;

	bool finished = true;

	while (iter.readID < iter.numofSeq) {

		//=============== pre filter package: 1 =======================================
		if(parameters.mode!=0){

			t1 = pre_filter_SW(liste, iter, reads_cpu, scaff_cpu, helper1, scaffold, parameters, finished);
		}else{
			t1 = pre_filter_SW_BS(liste, iter, reads_cpu, scaff_cpu, helper1, scaffold, parameters, finished,BSlookup);
		}
		//========================================================================
		amount = amount + t1;
		//=============== process result package: 2 =======================================
		if (t2 > 0) {
			post_proc_SW(results_cpu,reads,scaffold,results_gpu,parameters,t2,mem_result,helper2,rev);
		}
		//=======================================================================
		//=========================== call SW for package: 1 =================================

		if (t1 > 0) {

			CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

			if (parameters.mode == 0) {
				//calc SW BS mode
				if(rev){
					SW_CudaBS_AG<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
				} else {
					SW_CudaBS_TC<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
				}
			} else if(parameters.mode==1) {
				//calc SW normal
				SW_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
			} else if(parameters.mode==2) {
				Global_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
			}
		}
		//=====================================================================
		//=============== pre filter package: 2 =======================================
		if(t1==procs) {
			if(parameters.mode!=0){

				t2 =pre_filter_SW(liste,iter,reads_cpu, scaff_cpu,helper2,scaffold ,parameters,finished);
			}else{
				t2 =pre_filter_SW_BS(liste,iter,reads_cpu, scaff_cpu,helper2,scaffold ,parameters,finished,BSlookup);
			}
		}

		//		========================================================================
		amount = amount+t2;
		//=============== process result package:1 =======================================
		post_proc_SW(results_cpu,reads,scaffold,results_gpu, parameters,t1, mem_result,helper1,rev);

		//=======================================================================
		//=========================== call SW for package: 2 ====================================
		if(t2 > 0) {

			CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

			if(parameters.mode==0) {
				//calc SW BS mode
				if(rev) {
					SW_CudaBS_AG<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
				} else {
					SW_CudaBS_TC<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
				}
			} else if(parameters.mode==1) {
				//				calc SW normal
				SW_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
			} else if(parameters.mode==2) {
				Global_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
			}

		}
		//=======================================================================
	}

	if(t2>0) { //Just to be shure that I dont miss the last step!
		//=============== process result package: 2 if it is the end!: =======================================
		post_proc_SW(results_cpu,reads,scaffold,results_gpu, parameters,t2, mem_result,helper2,rev);
	}


	cout<<"\t\t\t"<<amount<<" where calculated"<<endl;

	delete [] parameters_cpu;
	delete [] results_cpu;
	helper1.clear();
	helper2.clear();
	delete [] reads_cpu;
	delete [] scaff_cpu;
	liste.clear();

	CUDA_SAFE_CALL(cudaFree(matrix_gpu));
	CUDA_SAFE_CALL(cudaFree(results_gpu));
	CUDA_SAFE_CALL(cudaFree(reads_gpu));
	CUDA_SAFE_CALL(cudaFree(scaffold_gpu));
	CUDA_SAFE_CALL(cudaFree(param));

	checkCUDAError("after delete");
}



void Sw_double(Reference & scaffold, Reads & reads, Parameters parameters, bool rev,vector< vector< vector< Match > > > & candidates) {

	long threads = 256; //fixed threads per block (number chosen to be the optimum)



	//calcualte the memory needed per thread:
	int size = reads.get_max_length() + parameters.corridor_with;
	long mem_parameters = 4;
	long mem_scores =4*parameters.match.size();


	long mem_per_thread = sizeof(float) * 2 * parameters.corridor_with + reads.get_max_length() * sizeof(char) + sizeof(float) + size * sizeof(char);
	//									matrix						     reads     					mem_result     scaff
	long mem2 = (parameters.memsize - mem_parameters * sizeof(int))-mem_scores* sizeof(float);

	//how many processors can we start?
	unsigned int blocks = (mem2 / (threads * mem_per_thread));
	blocks = blocks - parameters.mps;
	if (blocks > 2250) {
		blocks = 2250;
	}

	if (reads.get_max_length() > 200) {
		blocks = blocks / 2;
	}
	if (reads.get_max_length() > 700 && blocks > 300) {
		blocks = 300;
	}

	int procs = threads * blocks;
	//	cout<<procs<<endl;
	long mem_reads = reads.get_max_length() * procs;
	long mem_scaff = size * procs;
	long mem_result = procs*(int)parameters.match.size()*2;
	long mem_matrix = 2 * procs * parameters.corridor_with;

	char * reads_cpu = new char[mem_reads];
	char * scaff_cpu = new char[mem_scaff];
	float * results_cpu = new float[mem_result];
	int * parameters_cpu = new int[mem_parameters];
	float *scores_cpu = new float [mem_scores];


 	memset(results_cpu, -1.0f, mem_result * sizeof(float));

	//parameters needed on the GPU
	parameters_cpu[0] = (int)reads.get_max_length();
	parameters_cpu[1] = size;
	parameters_cpu[2] = (int)parameters.corridor_with; //corridor
	parameters_cpu[3] = (int)parameters.match.size();

	for(size_t i=0;i<parameters_cpu[3];i++){
		scores_cpu[4*i]=parameters.match[i];
		scores_cpu[1+4*i]=parameters.mismatch[i];
		scores_cpu[2+4*i]=parameters.gap_read[i];
		scores_cpu[3+4*i]=parameters.gap_scaffold[i];
	}
//	CUDA_SAFE_CALL(cudaMemcpyToSymbol(param, parameters_cpu,mem_parameters* sizeof(float)));

	int * param;
	CUDA_SAFE_CALL(cudaMalloc((void**)&param,mem_parameters*sizeof(int)));
	CUDA_SAFE_CALL(cudaMemcpy(param,parameters_cpu,mem_parameters*sizeof(int),cudaMemcpyHostToDevice));
	float *scores;
	CUDA_SAFE_CALL(cudaMalloc((void**)&scores,mem_scores*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemcpy(scores,scores_cpu,mem_scores*sizeof(float),cudaMemcpyHostToDevice));

	char * scaffold_gpu;
	char * reads_gpu;
	float * matrix_gpu;
	float * results_gpu;


	CUDA_SAFE_CALL(cudaMalloc((void**)&scaffold_gpu,mem_scaff*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&reads_gpu,mem_reads*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&results_gpu,mem_result*sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&matrix_gpu,mem_matrix*sizeof(float)));
	checkCUDAError("after mallocs");

	CUDA_SAFE_CALL(cudaMemset(matrix_gpu,0.0f,mem_matrix*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemset(results_gpu,-1.0f,mem_result*sizeof(float)));
	checkCUDAError("after memcpy");

	unsigned int shared_mem = (sizeof(char) + sizeof(unsigned int) + 2 * sizeof(float)) * threads;

	dim3 dimBlock(threads, 1);
	dim3 dimGrid(blocks, 1); //per matrix 1 thread!


	//needed for prefiltering:
	iterators_str iter;
	iter.readID = 0;
	iter.size = size;
	iter.seqlength = reads.get_max_length();
	iter.match = reads.matches.begin();
	iter.proc = procs;
	if (parameters.seedword_hits > 0) {
		iter.seed_tresh = parameters.seedword_hits;
	} else {
		iter.seed_tresh = call_parameter(reads.forward[0].size());
	}

	iter.shift = round((double) (parameters.corridor_with) / 2.0);
	iter.tresh = round((double) (parameters.corridor_with / 3.0));
	iter.numofSeq = reads.forward.size();

	if (reads.get_direction()) {
		iter.read_pos = reads.forward.begin();
		iter.mapps = reads.maped_reads.begin();
	} else if (!reads.get_direction()) {
		iter.read_pos = reads.backward.begin();
		iter.mapps = reads.maped_reads_back.begin();
	}

	int t1 = 0;
	int t2 = 0;

	vector<HelpStruct> helper1;
	vector<HelpStruct> helper2;

	HelpStruct temp;
	temp.id = 0;
	temp.position = 0;
	temp.scaff_id = 0;

	for (size_t i = 0; i <= procs; i++) {
		helper1.push_back(temp);
		helper2.push_back(temp);
	}

	map<int, list<ids_str> > liste;
	list<ids_str> tmp;
	for (int i = scaffold.get_ID(); i < scaffold.name.size(); i++) {
		liste[i] = tmp;
	}

	long amount = 0;

	bool finished = true;
	while (iter.readID < iter.numofSeq) {

		//=============== pre filter package: 1 =======================================
		t1 = pre_filter_SW(liste, iter, reads_cpu, scaff_cpu, helper1, scaffold, parameters, finished);
		//========================================================================
		amount = amount + t1;
		//=============== process result package: 2 =======================================
		if (t2 > 0) {
			post_proc_Double(results_cpu,scaffold,results_gpu,parameters,t2,mem_result,helper2,rev,candidates,reads.get_direction());
		}
		//=======================================================================
		//=========================== call SW for package: 1 =================================

		if (t1 > 0) {
			CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

//			if (parameters.mode == 0) {
//				//calc SW BS mode
//				if(rev){
//					SW_CudaBS_AG<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
//				} else {
//					SW_CudaBS_TC<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
//				}
//			}else[
			Score_Calc_Double<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param,scores);
//			}
		}
		//=====================================================================
		//=============== pre filter package: 2 =======================================
		if(t1==procs) {
			t2 =pre_filter_SW(liste,iter,reads_cpu, scaff_cpu,helper2,scaffold ,parameters,finished);
		}

		//		========================================================================
		amount = amount+t2;
		//=============== process result package:1 =======================================
		post_proc_Double(results_cpu,scaffold,results_gpu, parameters,t1, mem_result,helper1,rev,candidates,reads.get_direction());

		//=======================================================================
		//=========================== call SW for package: 2 ====================================
		if(t2 > 0) {
			CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

//			if(parameters.mode==0) {
//				//calc SW BS mode
//				if(rev) {
//					SW_CudaBS_AG<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
//				} else {
//					SW_CudaBS_TC<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
//				}
//			}else{
//				//				calc SW normal
			Score_Calc_Double<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param,scores);
//			}

		}
		//=======================================================================
	}

	if(t2>0) { //Just to be shure that I dont miss the last step!
		//=============== process result package: 2 if it is the end!: =======================================
		post_proc_Double(results_cpu,scaffold,results_gpu, parameters,t2, mem_result,helper2,rev,candidates,reads.get_direction());
	}

	cout<<"\t\t\t"<<amount*parameters.match.size()*2<<" where calculated"<<endl;

	delete [] parameters_cpu;
	delete [] results_cpu;
	helper1.clear();
	helper2.clear();
	delete [] reads_cpu;
	delete [] scaff_cpu;
	liste.clear();
	delete []scores_cpu;

	CUDA_SAFE_CALL(cudaFree(matrix_gpu));
	CUDA_SAFE_CALL(cudaFree(results_gpu));
	CUDA_SAFE_CALL(cudaFree(reads_gpu));
	CUDA_SAFE_CALL(cudaFree(scaffold_gpu));
	CUDA_SAFE_CALL(cudaFree(param));
	CUDA_SAFE_CALL(cudaFree(scores));

	checkCUDAError("after delete");
}

//TODO reconfigure!
void Sw_optimal(Reference & scaffold, Reads & reads, Parameters parameters, bool rev) {

	long threads = 256; //fixed threads per block (number chosen to be the optimum)

	int size = reads.get_max_length() + parameters.corridor_with;

	//calcualte the memory needed per thread:

	long mem_parameters = 9;
	long blocks = 2250;

	if (reads.get_max_length() > 200) {
		blocks = blocks / 2;
	}
	if (reads.get_max_length() > 700 && blocks > 300) {
		blocks = 300;
	}

	int procs = threads * blocks;
	cout << "procs " << procs << endl;

	long mem_reads = reads.get_max_length() * procs;
	long mem_score = procs;
	long mem_positions = procs;
	long mem_matrix = 2 * procs * reads.get_max_length();

	char * reads_cpu = new char[mem_reads];
	float * score_cpu = new float[mem_score];
	int * position_cpu = new int[mem_positions];
	float * parameters_cpu = new float[mem_parameters];

	memset(score_cpu, -1.0f, mem_score * sizeof(float));

	//parameters needed on the GPU
	parameters_cpu[0] = (float) reads.get_max_length();

	parameters_cpu[2] = parameters.match[0];
	parameters_cpu[3] = parameters.mismatch[0];
	parameters_cpu[4] = parameters.gap_read[0];
	//	parameters_cpu[5] =(float)parameters.corridor_with; //corridor
	parameters_cpu[6] = parameters.gap_scaffold[0];
	parameters_cpu[7] = parameters.T_match[0];
	parameters_cpu[8] = parameters.T_mismatch[0];

	char * scaffold_gpu;
	char * reads_gpu;
	float * matrix_gpu;
	int * positions_gpu;
	float * score_gpu;
	float *param;
	CUDA_SAFE_CALL(cudaMalloc((void**)&param,mem_parameters*sizeof(float)));

	CUDA_SAFE_CALL(cudaMalloc((void**)&reads_gpu,mem_reads*sizeof(char)));

	CUDA_SAFE_CALL(cudaMalloc((void**)&positions_gpu,mem_positions*sizeof(int)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&score_gpu,mem_score*sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&matrix_gpu,mem_matrix*sizeof(float)));

	checkCUDAError("after mallocs");

	CUDA_SAFE_CALL(cudaMemset(matrix_gpu,0.0f,mem_matrix*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemset(score_gpu,-1.0f,mem_score*sizeof(float)));
	checkCUDAError("after memcpy");

	unsigned int shared_mem = (sizeof(char) + sizeof(unsigned int) + 2 * sizeof(float)) * threads;

	dim3 dimBlock(threads, 1);
	dim3 dimGrid(blocks, 1); //per matrix 1 thread!


	iterators_str iter;
	iter.readID = 0;
	iter.numofSeq = reads.forward.size();
	iter.size = size;
	iter.seqlength = reads.get_max_length();
	iter.match = reads.matches.begin();
	iter.proc = procs;

	if (reads.get_direction()) {
		iter.read_pos = reads.forward.begin();
	} else if (!reads.get_direction()) {
		iter.read_pos = reads.backward.begin();
	}
	vector<int> helper1;
	vector<int> helper2;

	for (size_t i = 0; i < procs; i++) {
		helper1.push_back(-1);
		helper2.push_back(-1);
	}

	checkCUDAError("after mallocs");
	cout << "start" << endl;

	size_t t1 = 0;
	size_t t2 = 0;
	cout << scaffold.sequence[0].size() << endl;

	//Think about smaller genomes!!
	for (size_t scaff_id = 0; scaff_id < scaffold.sequence.size(); scaff_id++) {

		CUDA_SAFE_CALL(cudaMalloc((void**)&scaffold_gpu,scaffold.sequence[scaff_id].size()*sizeof(char)));
		checkCUDAError("after alloc scaff");
		parameters_cpu[1] = (float) scaffold.sequence[scaff_id].size();
		CUDA_SAFE_CALL(cudaMemcpy(param,parameters_cpu,mem_parameters*sizeof(float),cudaMemcpyHostToDevice));
//		CUDA_SAFE_CALL(cudaMemcpyToSymbol(param, parameters_cpu,mem_parameters* sizeof(float)));

		checkCUDAError("after param");

		CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaffold.sequence[scaff_id].c_str(),scaffold.sequence[scaff_id].size()*sizeof(char),cudaMemcpyHostToDevice));

		checkCUDAError("after mallocs scaff");

		while (iter.readID < iter.numofSeq) {
			//=============== pre filter package: 1 =======================================
			cout << "pre 1" << endl;
			t1 = pre_filter_SW_optimal(reads, reads_cpu, iter, helper1);
			//========================================================================
			//========================================================================
			if (t2 > 0) {
				cout << "post 2" << endl;
				post_proc_SW_optimal(score_gpu, score_cpu, positions_gpu, position_cpu, reads, scaffold, t2, mem_score, mem_positions, scaff_id, helper2, iter, rev);
				//				post_proc_SW(score_gpu,scores_cpu,positoin_gpu,positions_cpu,reads,ref,t2,mem_score,mem_positions);
			}
			//=======================================================================
			//=========================== call SW for package: 1 =================================
			if (t1 > 0) {
				cout << "call 1" << endl;
				CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

				//reads are copied in pre filtering
				SW_Cuda_optimal<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,score_gpu,positions_gpu,param);
				checkCUDAError(" end call1");

			}
			//=====================================================================
			//=============== pre filter package: 2 =======================================
			if (t1 == procs) {
				cout << "pre 2" << endl;
				t2 = pre_filter_SW_optimal(reads, reads_cpu, iter, helper2);
			}

			//========================================================================
			//=============== process result package:1 =======================================
			cout << "post 1" << endl;
			post_proc_SW_optimal(score_gpu, score_cpu, positions_gpu, position_cpu, reads, scaffold, t1, mem_score, mem_positions, scaff_id, helper1, iter, rev);

			//=======================================================================
			//=========================== call SW for package: 2 ====================================
			if (t2 > 0) {
				cout << "call 2" << endl;
				CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

				SW_Cuda_optimal<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,score_gpu,positions_gpu,param);
				checkCUDAError(" call2 ");

			}
			//=======================================================================
		}

		if (t2 > 0) { //Just to be shure that I dont miss the last step!
			//=============== process result package: 2 if it is the end!: =======================================

			post_proc_SW_optimal(score_gpu, score_cpu, positions_gpu, position_cpu, reads, scaffold, t1, mem_score, mem_positions, scaff_id, helper2, iter, rev);
		}
		CUDA_SAFE_CALL(cudaFree(scaffold_gpu));

	}

	cout << "\t\t\t where calculated" << endl;

	delete[] parameters_cpu;
	delete[] score_cpu;
	helper1.clear();
	helper2.clear();
	delete[] reads_cpu;

	CUDA_SAFE_CALL(cudaFree(matrix_gpu));
	CUDA_SAFE_CALL(cudaFree(positions_gpu));
	CUDA_SAFE_CALL(cudaFree(reads_gpu));
	CUDA_SAFE_CALL(cudaFree(score_gpu));
	CUDA_SAFE_CALL(cudaFree(param));

	checkCUDAError("after delete");
}




bool check(vector<Match>::iterator null,vector<Match>::iterator curr){
	return ((*curr).position==(*null).position &&(*null).forward ==(*curr).forward && (*null).BS_rev == (*curr).BS_rev && (*null).scaffID==(*curr).scaffID);
}

void CalcOutput3(vector<string> scaffold_file, Reads &reads, Parameters parameters) {

	//	int id= 3;
	//	cout<<reads.matches[id].match[0].score<<endl;
	unsigned int threads = 256;
	int size = parameters.corridor_with + reads.get_max_length();
	int parameter_mem = 9;

	//checking mem:
	long mem_per_thread = size * sizeof(char) + reads.get_max_length() * sizeof(char) + size * 2 * sizeof(char) + parameters.corridor_with * reads.get_max_length() * sizeof(float) + 3 * sizeof(int);
									//scaff   		             read					       result									mat							                        pos
	long mem2 = parameters.memsize - (parameter_mem * sizeof(float));

	unsigned int blocks = (mem2 / (threads * mem_per_thread));
	blocks = blocks - (blocks % parameters.mps);
	blocks = blocks - parameters.mps;

	if (blocks > 2250 / 2) {
		blocks = 2250 / 2;
	}

	if (reads.get_max_length() > 200) {
		blocks = blocks / 2;
	}
	if (reads.get_max_length() > 700 && blocks > 300) {
		blocks = 300;
	}

	long procs = threads * blocks;


	//  mem size:
	long mem_scaffold = size * procs;
	long mem_reads = reads.get_max_length() * procs;
	long mem_result = size * procs * 2;
	long mem_matrix = (long)parameters.corridor_with * (long)reads.get_max_length() * procs;
	long mem_pos = 2 * procs;
	long mem_matches = procs;
	long mem_rev;

	if (parameters.mode == 0) {
		mem_rev = procs;
	} else {
		mem_rev = 0;
	}

	char * scaffold_cpu = new char[mem_scaffold];
	char * reads_cpu = new char[mem_reads];
	char * result_cpu = new char[mem_result];
	float * parameters_cpu = new float[parameter_mem];
	int * pos_cpu = new int[mem_pos];
	float * matches_cpu = new float[mem_matches];
	bool * rev_cpu = new bool[mem_rev];

	char * scaffold_gpu;
	char * reads_gpu;
	char * results_gpu;
	float * parameters_gpu;
	float * matrix_gpu;
	int * pos_gpu;
	float * matches_gpu;
	bool * rev_gpu;

	CUDA_SAFE_CALL(cudaMalloc((void**)&scaffold_gpu,mem_scaffold*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&reads_gpu,mem_reads*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&results_gpu,mem_result*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&parameters_gpu,parameter_mem*sizeof(float)));

	CUDA_SAFE_CALL(cudaMalloc((void**)&matrix_gpu,mem_matrix*sizeof(float)));
	checkCUDAError("malloc 1");
	CUDA_SAFE_CALL(cudaMalloc((void**)&pos_gpu,mem_pos*sizeof(int)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&matches_gpu,mem_matches*sizeof(float)));

	CUDA_SAFE_CALL(cudaMalloc((void**)&rev_gpu,mem_rev*sizeof(bool)));
	checkCUDAError("malloc 1");

	parameters_cpu[0] = (float) reads.get_max_length();
	parameters_cpu[1] = (float) size;
	parameters_cpu[2] = parameters.match[0]; //match
	parameters_cpu[3] = parameters.mismatch[0]; //missmatch
	parameters_cpu[4] = parameters.gap_read[0]; //gap
	parameters_cpu[5] = (float) parameters.corridor_with; //corridor
	parameters_cpu[6] = parameters.gap_scaffold[0]; //gap scaffold
	parameters_cpu[7] = parameters.T_match[0];
	parameters_cpu[8] = parameters.T_mismatch[0];

	int shift = (int)round((double) (parameters.corridor_with) / 2.0);

//	CUDA_SAFE_CALL(cudaMemcpyToSymbol(param, parameters_cpu,mem_parameters* sizeof(float)));

	float * param;
	CUDA_SAFE_CALL(cudaMalloc((void**)&param,parameter_mem*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemcpy(param,parameters_cpu,parameter_mem*sizeof(float),cudaMemcpyHostToDevice));

	unsigned int shared_mem = (sizeof(float) + 2 * sizeof(unsigned int)) * threads;

	dim3 dimBlock(threads, 1);
	dim3 dimGrid(blocks, 1); //per matrix 1 thread!

	CUDA_SAFE_CALL(cudaMemset(matrix_gpu,0.0f,mem_matrix*sizeof(float)));
	checkCUDAError("memcpy");

	int counter = 0;
	int * submitted = new int[2 * (procs)];

	Reference scaffold((reads.get_max_length() + (parameters.corridor_with)));

	vector< size_t> dels;
	for (vector<string>::iterator ref = scaffold_file.begin(); ref != scaffold_file.end(); ref++) { //loop over files
		scaffold.set_filename((*ref));

		do {//loop over file
			// Read in Chromosome:
			getNextScaff(scaffold, parameters);
			if (scaffold.sequence[0].size() > 0) {
				for (size_t i = 0; i < reads.forward.size(); i++) {
				 	int position = 0;
					for (vector<Match>::iterator j = reads.matches[i].match.begin(); j != reads.matches[i].match.end(); j++) {
						if ((*j).scaffID >= scaffold.get_ID() && (*j).scaffID < scaffold.name.size()) {

							if (parameters.mode == 2 || (*j).score > parameters.score_treshold) { //within this piece

								(*j).position -= shift;
								//copy
								submitted[counter * 2] = i;
								submitted[1 + counter * 2] = position;

								//Needed for AG mismatches in BS mode
								if (parameters.mode == 0) {
									rev_cpu[counter] = (*j).BS_rev;
								}

								if ((*j).forward) {
									memcpy(&reads_cpu[counter * reads.get_max_length()], reads.forward[i].c_str(), reads.get_max_length() * sizeof(char));
								} else if (!(*j).forward) {
									memcpy(&reads_cpu[counter * reads.get_max_length()], reads.backward[i].c_str(), reads.get_max_length() * sizeof(char));
								}



								//copy the scaffold pieces together:
								int pos = (*j).position - scaffold.get_scaff_pos((*j).scaffID);

								if (pos > 0 && pos + size < scaffold.sequence[(*j).scaffID - scaffold.get_ID()].size()) {
									memcpy(&scaffold_cpu[size * counter], scaffold.sequence[(*j).scaffID - scaffold.get_ID()].substr((size_t) pos, (size_t) size).c_str(), size * sizeof(char));
								} else {
									int z = 0; //size*counter;
									for (int t = pos; t < size + pos; t++) {
										if (t < 0 || t >= scaffold.sequence[(*j).scaffID - scaffold.get_ID()].size()) {
											scaffold_cpu[z + size * counter] = 'I';
										} else {
											scaffold_cpu[z + size * counter] = scaffold.sequence[(*j).scaffID - scaffold.get_ID()][t];
										}

										z++;
									}

								}

								//copy the read sequences together:

								counter++;
							}

							if (counter == procs) { //enough jobs sampled
								//submit job
								CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaffold_cpu,mem_scaffold*sizeof(char),cudaMemcpyHostToDevice));
								CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));
								CUDA_SAFE_CALL(cudaMemset(results_gpu,' ',mem_result*sizeof(char)));
								CUDA_SAFE_CALL(cudaMemset(pos_gpu,0,mem_pos*sizeof(int)));
								CUDA_SAFE_CALL(cudaMemcpy(rev_gpu,rev_cpu,mem_rev*sizeof(bool),cudaMemcpyHostToDevice));

								if (parameters.mode == 0) {
									SW_CudaBS_Back2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,parameters_gpu,pos_gpu,matches_gpu,rev_gpu,param);
								} else if(parameters.mode==1) {
									SW_Cuda_Back2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,parameters_gpu,pos_gpu,matches_gpu,param);
								} else if(parameters.mode==2 || parameters.mode==4 ) {
									Global_Cuda_Back2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,parameters_gpu,pos_gpu,matches_gpu,param);
								}
								//copy results
								CUDA_SAFE_CALL(cudaMemcpy(pos_cpu,pos_gpu,mem_pos*sizeof(int),cudaMemcpyDeviceToHost));
								CUDA_SAFE_CALL(cudaMemcpy(result_cpu,results_gpu,mem_result*sizeof(char),cudaMemcpyDeviceToHost));
								CUDA_SAFE_CALL(cudaMemcpy(matches_cpu,matches_gpu,mem_matches*sizeof(float),cudaMemcpyDeviceToHost));

								//store results per read
								for(int t=0;t<counter;t++) {
									vector<Match>::iterator match=reads.matches[submitted[t*2]].match.begin()+submitted[1+t*2];

									(*match).position+= pos_cpu[t*2];
									if(submitted[1+t*2]!=0 && check(reads.matches[submitted[t*2]].match.begin(),match)){
										dels.push_back(submitted[t*2]);
									}


									float count_length=0.0f;
									float count_length2=0.0f;
									for(int z=0;z<size;z++) {

										if(result_cpu[z+t*size*2]=='#') { //if fill up char is detected (read ends here)
											break;
										}
										if(result_cpu[z+t*size*2]!=' ') {
											(*match).read+=result_cpu[z+t*size*2]; //stores the read sequence
											if(result_cpu[z+t*size*2]!='-' ) {
												count_length2++;
											 if(result_cpu[z+size+t*size*2]!='-'){
												 count_length++;
												}
											}

											(*match).scaff+=result_cpu[z+size+t*size*2]; //stores the reference sequence
										}
									}

									(*match).identity=matches_cpu[t]/count_length;
									count_length2+=pos_cpu[1+t*2];

									(*match).Q_start=pos_cpu[1+t*2];
									(*match).Q_end=reads.matches[submitted[t*2]].orig_length-count_length2;

								}


								counter=0;
							}
						}
						position++;
					}
				}
			}
		}while(scaffold.sequence[0].size()!=0);
	}

	if(counter>0) { //if somethings are left to process
		int rest = (procs)-counter;
		memset(&scaffold_cpu[size*counter],' ',rest*size*sizeof(char));
		memset(&reads_cpu[reads.get_max_length()*counter],' ',rest*reads.get_max_length()*sizeof(char));
		memset(&submitted[counter*2],-1,rest*2*sizeof(int));

		CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaffold_cpu,mem_scaffold*sizeof(char),cudaMemcpyHostToDevice));
		CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));
		CUDA_SAFE_CALL(cudaMemset(results_gpu,' ',mem_result*sizeof(char)));
		CUDA_SAFE_CALL(cudaMemset(pos_gpu,0,mem_pos*sizeof(int)));
		CUDA_SAFE_CALL(cudaMemcpy(rev_gpu,rev_cpu,mem_rev*sizeof(bool),cudaMemcpyHostToDevice));

		if(parameters.mode==0) {
			SW_CudaBS_Back2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,parameters_gpu,pos_gpu,matches_gpu,rev_gpu,param);
		} else if(parameters.mode==1) {
			SW_Cuda_Back2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,parameters_gpu,pos_gpu,matches_gpu,param);
		} else if(parameters.mode==2|| parameters.mode==4 ) {
			Global_Cuda_Back2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,parameters_gpu,pos_gpu,matches_gpu,param);
		}

		CUDA_SAFE_CALL(cudaMemcpy(pos_cpu,pos_gpu,mem_pos*sizeof(int),cudaMemcpyDeviceToHost));
		CUDA_SAFE_CALL(cudaMemcpy(result_cpu,results_gpu,mem_result*sizeof(char),cudaMemcpyDeviceToHost));
		CUDA_SAFE_CALL(cudaMemcpy(matches_cpu,matches_gpu,mem_matches*sizeof(float),cudaMemcpyDeviceToHost));

		//process results (look above)
		for(int t=0;t<counter;t++) {

			vector<Match>::iterator match=reads.matches[submitted[t*2]].match.begin()+submitted[1+t*2];

			(*match).position+= pos_cpu[t*2];
			if(submitted[1+t*2]!=0 && check(reads.matches[submitted[t*2]].match.begin(),match)){
				dels.push_back(submitted[t*2]);
			}
			float count_length=0.0f;
			float count_length2=0.0f;

			for(int z=0;z<size;z++) {
				if(result_cpu[z+t*size*2]=='#') {
					break;
				}
				if(result_cpu[z+t*size*2]!=' ') {
					(*match).read+=result_cpu[z+t*size*2];
					if(result_cpu[z+t*size*2]!='-' ) {
						count_length2++;
					 if(result_cpu[z+size+t*size*2]!='-'){
						 count_length++;
						}
					}
					(*match).scaff+=result_cpu[z+size+t*size*2];
				}
			}

			(*match).identity=matches_cpu[t]/count_length;
			count_length2+=pos_cpu[1+t*2];
			(*match).Q_start=pos_cpu[1+t*2];
			(*match).Q_end=reads.matches[submitted[t*2]].orig_length-count_length2;

		}

	}

	for(size_t j=0;j<dels.size();j++){
		reads.matches[dels[j]].match.erase(reads.matches[dels[j]].match.begin());
	}

	delete [] parameters_cpu;
	delete [] submitted;
	delete [] scaffold_cpu;
	delete [] reads_cpu;
	delete [] result_cpu;
	delete [] pos_cpu;
	delete [] matches_cpu;
	delete [] rev_cpu;

	CUDA_SAFE_CALL(cudaFree(parameters_gpu));
	CUDA_SAFE_CALL(cudaFree(matrix_gpu));
	CUDA_SAFE_CALL(cudaFree(reads_gpu));
	CUDA_SAFE_CALL(cudaFree(results_gpu));
	CUDA_SAFE_CALL(cudaFree(scaffold_gpu));
	CUDA_SAFE_CALL(cudaFree(pos_gpu));
	CUDA_SAFE_CALL(cudaFree(matches_gpu));
	CUDA_SAFE_CALL(cudaFree(rev_gpu));
	CUDA_SAFE_CALL(cudaFree(param));

}


vector<HashKey> hash_read(string seq, Parameters parameters){
	vector<HashKey> mapped;
	 for (size_t j = 0; j + parameters.wordsize < seq.size(); j = j + parameters.step_size) {
		HashKey temp;
		//if not A,C,G,T FatsHash returns -1. -> will not be stored
		temp.key= FatsHash2(seq.substr(j,parameters.wordsize));//,parameters);

		if (temp.key >= 0 && temp.key <= parameters.mapsize) {
			temp.pos = j;
			mapped.push_back(temp);
		}
	}
	return mapped;
 }
void CalcPairs(vector<string> scaffold_file, Reads & pairs, Parameters parameters, Reads reads) {
	long threads = 256; //fixed threads per block (number chosen to be the optimum)


	//calcualte the memory needed per thread:
	int size = reads.get_max_length() + parameters.corridor_with;
	long mem_parameters = 9;
	long mem_per_thread = sizeof(float) * 2 * parameters.corridor_with + reads.get_max_length() * sizeof(char) + sizeof(float) + size * sizeof(char);
	//									matrix						                                	reads     					mem_result     scaff
	long mem2 = parameters.memsize - (mem_parameters * sizeof(float));

	//how many processors can we start?
	long blocks = (mem2 / (threads * mem_per_thread));
	blocks = blocks - parameters.mps;
	if (blocks > 2250) {
		blocks = 2250;
	}

	if (reads.get_max_length() > 200) {
		blocks = blocks / 2;
	}
	if (reads.get_max_length() > 700 && blocks > 300) {
		blocks = 300;
	}

	long procs = threads * blocks;
	long mem_reads = reads.get_max_length() * procs;
	long mem_scaff = size * procs;
	long mem_result = procs;
	long mem_matrix = 2 * procs * parameters.corridor_with;

	char * reads_cpu = new char[mem_reads];
	char * scaff_cpu = new char[mem_scaff];
	float * results_cpu = new float[mem_result];
	float * parameters_cpu = new float[mem_parameters];

	memset(results_cpu, -1.0f, mem_result * sizeof(float));

	//parameters needed on the GPU
	parameters_cpu[0] = (float) reads.get_max_length();
	parameters_cpu[1] = (float) size;
	parameters_cpu[2] = parameters.match[0];
	parameters_cpu[3] = parameters.mismatch[0];
	parameters_cpu[4] = parameters.gap_read[0];
	parameters_cpu[5] = (float) parameters.corridor_with; //corridor
	parameters_cpu[6] = parameters.gap_scaffold[0];
	parameters_cpu[7] = parameters.T_match[0];
	parameters_cpu[8] = parameters.T_mismatch[0];

//	CUDA_SAFE_CALL(cudaMemcpyToSymbol(param, parameters_cpu,mem_parameters* sizeof(float)));

	float * param;
	CUDA_SAFE_CALL(cudaMalloc((void**)&param,mem_parameters*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemcpy(param,parameters_cpu,mem_parameters*sizeof(float),cudaMemcpyHostToDevice));


	char * scaffold_gpu;
	char * reads_gpu;
	float * matrix_gpu;
	float * results_gpu;

	CUDA_SAFE_CALL(cudaMalloc((void**)&scaffold_gpu,mem_scaff*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&reads_gpu,mem_reads*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&results_gpu,mem_result*sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&matrix_gpu,mem_matrix*sizeof(float)));
	checkCUDAError("after mallocs");

	CUDA_SAFE_CALL(cudaMemset(matrix_gpu,0.0f,mem_matrix*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemset(results_gpu,-1.0f,mem_result*sizeof(float)));
	checkCUDAError("after memcpy");

	unsigned int shared_mem = (sizeof(char) + sizeof(unsigned int) + 2 * sizeof(float)) * threads;

	dim3 dimBlock(threads, 1);
	dim3 dimGrid(blocks, 1); //per matrix 1 thread!


	Reference scaffold(parameters.mapsize,(reads.get_max_length()+(parameters.corridor_with))); //init

	list< pair <int,int> > liste;
	long t1 = 0;

	vector<HelpStruct_pair> helper1;
	HelpStruct_pair temp;
	temp.id = 0;
	temp.position = 0;
	temp.scaff_id = 0;


	for (size_t i = 0; i <= procs; i++) {
		helper1.push_back(temp);
	}

	iterators_pair_str iter;
	iter.proc = procs;

	iter.seqlength = pairs.get_max_length();
	iter.size = size;
	iter.shift = round((double) (parameters.corridor_with) / 2.0);
	iter.tresh = round((double) (parameters.corridor_with / 3.0));
	long amount = 0;



	for (vector<string>::iterator ref = scaffold_file.begin(); ref != scaffold_file.end(); ref++) { //loop over files
		scaffold.set_filename((*ref));

		do {//loop over file
			// Read in Chromosome:
			getNextScaff(scaffold, parameters);

			if (scaffold.sequence[0].size() > 0) {
				scaffold.map_key(parameters);

				for (size_t i = 0; i < reads.matches.size(); i++) {
					iter.seqlength = pairs.get_max_length();
					iter.read_ID = i;


					for (vector<Match>::iterator match = reads.matches[i].match.begin(); match != reads.matches[i].match.end(); match++) {

						if ((*match).identity > parameters.identity_treshold && (*match).score > parameters.score_treshold) {
							iter.match = match;

							//check if the ref name is the same! not the scaffID
							for (size_t scaff_pos = scaffold.get_ID(); scaff_pos < scaffold.name.size(); scaff_pos++) {

								if (strcmp(scaffold.get_name(scaff_pos).c_str(), scaffold.get_name((*match).scaffID).c_str()) == 0) {
									//check the scaff ID according if the pair end should be positioned infornt or behind the read
									if (i < pairs.matches.size()) {

										//TODO check if this works also for BS!
										iter.forward=!(*match).forward; //Because of the paired end theory
										iter.BS_rev = (*match).BS_rev;
										iter.scaffID = scaff_pos;

										if (!iter.forward &&(iter.scaffID >= (*iter.match).scaffID) ) {

											//if first read matched forward second have to be backward
											vector<HashKey> mapped =hash_read(pairs.backward[iter.read_ID],parameters);
											iter.key_begin = mapped.begin();
											iter.key_end = mapped.end();
											iter.seq = pairs.backward[iter.read_ID];

											pre_filter_SW_pair(t1, liste, iter, helper1, reads_cpu, scaff_cpu, scaffold, parameters);
											//The read have to be behind the other read part

											mapped.clear();

										} else if (iter.scaffID <= (*iter.match).scaffID){
											//if first read matched backward second have to be forward

											vector<HashKey> mapped =hash_read(pairs.forward[iter.read_ID],parameters);
											iter.key_begin = mapped.begin();
											iter.key_end = mapped.end();
											iter.seq = pairs.forward[iter.read_ID];
											pre_filter_SW_pair(t1, liste, iter, helper1, reads_cpu, scaff_cpu, scaffold, parameters);
											//The read have to be infornt the other read part

										}
									}
									if (t1 == procs) {
										while (t1 == procs) {
											amount += t1;
											//calc alignments
											CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
											CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

											if (parameters.mode == 1) {
												//				calc SW normal
												SW_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
											} else if(parameters.mode==2) {
												Global_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
											}

											//store results.
											//copy results
											post_proc_pair(helper1,pairs,t1,results_cpu,results_gpu,mem_result,reads,scaffold,parameters.corridor_with);

											//copy the left overs
											pre_filter_SW_pair_rest(t1,liste,iter,helper1,reads_cpu,scaff_cpu,scaffold,parameters);
										}
									}
								}
							}
//							iter.match_id++;
						}
					}
				}
				scaffold.clear_scaff_keys();
			}
		}while(scaffold.sequence[0].size()!=0);
	}
	if(t1!=0) {
		amount+=t1;
		memset(&reads_cpu[t1*iter.seqlength],' ',(iter.proc-t1)*iter.seqlength*sizeof(char));
		//calc alignments
		CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
		CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));
		checkCUDAError("end2 1");

		if(parameters.mode==1) {
		//	calc SW normal
			SW_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
		} else if(parameters.mode==2) {
			Global_Cuda2<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param);
		}
		//		checkCUDAError("end2 2");
		//store results.
		//copy results
		post_proc_pair(helper1,pairs,t1, results_cpu,results_gpu,mem_result,reads,scaffold,parameters.corridor_with);
	}
	cout<<"\t\t\t"<<amount<<" where calculated"<<endl;

	delete [] parameters_cpu;
	delete [] results_cpu;
	helper1.clear();

	delete [] reads_cpu;
	delete [] scaff_cpu;
	liste.clear();

	CUDA_SAFE_CALL(cudaFree(matrix_gpu));
	CUDA_SAFE_CALL(cudaFree(results_gpu));
	CUDA_SAFE_CALL(cudaFree(reads_gpu));
	CUDA_SAFE_CALL(cudaFree(scaffold_gpu));
	CUDA_SAFE_CALL(cudaFree(param));

	checkCUDAError("after delete");
}








void CalcPairs_BS(vector<string> scaffold_file, Reads & pairs, Parameters parameters, Reads reads) {
	long threads = 256; //fixed threads per block (number chosen to be the optimum)
	cout<<"BS_pair"<<endl;

	//calcualte the memory needed per thread:
	int size = reads.get_max_length() + parameters.corridor_with;
	long mem_parameters = 9;
	long mem_per_thread = sizeof(float) * 2 * parameters.corridor_with + reads.get_max_length() * sizeof(char) + sizeof(float) + size * sizeof(char);
	//									matrix						                                	reads     					mem_result     scaff
	long mem2 = parameters.memsize - (mem_parameters * sizeof(float));

	//how many processors can we start?
	long blocks = (mem2 / (threads * mem_per_thread));
	blocks = blocks - parameters.mps;
	if (blocks > 2250) {
		blocks = 2250;
	}

	if (reads.get_max_length() > 200) {
		blocks = blocks / 2;
	}
	if (reads.get_max_length() > 700 && blocks > 300) {
		blocks = 300;
	}

	long procs = threads * blocks;

	long mem_reads = reads.get_max_length() * procs;
	long mem_scaff = size * procs;
	long mem_result = procs;
	long mem_matrix = 2 * procs * parameters.corridor_with;
	long mem_bs= procs;

	char * reads_cpu = new char[mem_reads];
	char * scaff_cpu = new char[mem_scaff];
	float * results_cpu = new float[mem_result];
	float * parameters_cpu = new float[mem_parameters];
	bool * bs_cpu = new bool [mem_bs];

	memset(results_cpu, -1.0f, mem_result * sizeof(float));

	//parameters needed on the GPU
	parameters_cpu[0] = (float) reads.get_max_length();
	parameters_cpu[1] = (float) size;
	parameters_cpu[2] = parameters.match[0];
	parameters_cpu[3] = parameters.mismatch[0];
	parameters_cpu[4] = parameters.gap_read[0];
	parameters_cpu[5] = (float) parameters.corridor_with; //corridor
	parameters_cpu[6] = parameters.gap_scaffold[0];
	parameters_cpu[7] = parameters.T_match[0];
	parameters_cpu[8] = parameters.T_mismatch[0];

//	CUDA_SAFE_CALL(cudaMemcpyToSymbol(param, parameters_cpu,mem_parameters* sizeof(float)));

	float * param;
	CUDA_SAFE_CALL(cudaMalloc((void**)&param,mem_parameters*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemcpy(param,parameters_cpu,mem_parameters*sizeof(float),cudaMemcpyHostToDevice));


	bool * bs_gpu;
	char * scaffold_gpu;
	char * reads_gpu;
	float * matrix_gpu;
	float * results_gpu;

	CUDA_SAFE_CALL(cudaMalloc((void**)&scaffold_gpu,mem_scaff*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&reads_gpu,mem_reads*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&results_gpu,mem_result*sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&matrix_gpu,mem_matrix*sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&bs_gpu,mem_bs*sizeof(bool)));
	checkCUDAError("after mallocs");

	CUDA_SAFE_CALL(cudaMemset(matrix_gpu,0.0f,mem_matrix*sizeof(float)));
	CUDA_SAFE_CALL(cudaMemset(results_gpu,-1.0f,mem_result*sizeof(float)));
	checkCUDAError("after memcpy");

	unsigned int shared_mem = (sizeof(char) + sizeof(unsigned int) + 2 * sizeof(float)) * threads;

	dim3 dimBlock(threads, 1);
	dim3 dimGrid(blocks, 1); //per matrix 1 thread!

	Reference scaffold(parameters.mapsize,(reads.get_max_length()+(parameters.corridor_with))); //init

	list< pair <int,int> > liste;
	long t1 = 0;

	vector<HelpStruct_pair> helper1;
	HelpStruct_pair temp;
	temp.id = 0;
	temp.position = 0;
	temp.scaff_id = 0;


	for (size_t i = 0; i <= procs; i++) {
		helper1.push_back(temp);
	}

	iterators_pair_str iter;
	iter.proc = procs;

	iter.seqlength = pairs.get_max_length();
	iter.size = size;
	iter.shift = round((double) (parameters.corridor_with) / 2.0);
	iter.tresh = round((double) (parameters.corridor_with / 3.0));
	long amount = 0;


	vector< vector< int > > BSlookupTC,BSlookupAG ;
	BSlookupTC=parseLookup(parameters,false);
	BSlookupAG=parseLookup(parameters,true);

	for (vector<string>::iterator ref = scaffold_file.begin(); ref != scaffold_file.end(); ref++) { //loop over files
		scaffold.set_filename((*ref));

		do {//loop over file
			// Read in Chromosome:
			getNextScaff(scaffold, parameters);

			if (scaffold.sequence[0].size() > 0) {

				scaffold.map_key(parameters);
				for (size_t i = 0; i < reads.matches.size(); i++) {
					iter.seqlength = pairs.get_max_length();
					iter.read_ID = i;

					for (vector<Match>::iterator match = reads.matches[i].match.begin(); match != reads.matches[i].match.end(); match++) {

						if ((*match).identity > parameters.identity_treshold && (*match).score > parameters.score_treshold) {
							iter.match = match;

							//check if the ref name is the same! not the scaffID
							for (size_t scaff_pos = scaffold.get_ID(); scaff_pos < scaffold.name.size(); scaff_pos++) {

								if (strcmp(scaffold.get_name(scaff_pos).c_str(), scaffold.get_name((*match).scaffID).c_str()) == 0) {
									//check the scaff ID according if the pair end should be positioned in front or behind the read
									if (i < pairs.matches.size()) {

										iter.forward=!(*match).forward; //Because of the paired end theory
										iter.scaffID = scaff_pos;
										iter.BS_rev=(*match).BS_rev;


										//The read have to be behind the other read part
										if (!iter.forward &&iter.scaffID >= (*iter.match).scaffID ) {
											vector<HashKey> mapped =hash_read(pairs.backward[iter.read_ID],parameters);

											iter.key_begin = mapped.begin();
											iter.key_end = mapped.end();
											//if first read matched forward second have to be backward
											iter.seq = pairs.backward[iter.read_ID];
											if(iter.BS_rev){
												pre_filter_SW_pair(t1, liste, iter, helper1, reads_cpu, scaff_cpu, scaffold, parameters,BSlookupAG,bs_cpu);
											}else{
												pre_filter_SW_pair(t1, liste, iter, helper1, reads_cpu, scaff_cpu, scaffold, parameters,BSlookupTC,bs_cpu);
											}
											mapped.clear();

										} else if (iter.scaffID <= (*iter.match).scaffID){
											//if first read matched backward second have to be forward

											vector<HashKey> mapped =hash_read(pairs.forward[iter.read_ID],parameters);

											iter.key_begin = mapped.begin();
											iter.key_end = mapped.end();
											iter.seq = pairs.forward[iter.read_ID];
											if(iter.BS_rev){
												pre_filter_SW_pair(t1, liste, iter, helper1, reads_cpu, scaff_cpu, scaffold, parameters,BSlookupAG,bs_cpu);
											}else{
												pre_filter_SW_pair(t1, liste, iter, helper1, reads_cpu, scaff_cpu, scaffold, parameters,BSlookupTC,bs_cpu);
											}
											//The read have to be in front the other read part
											mapped.clear();
										}
									}
									if (t1 == procs) {
										while (t1 == procs) {
											amount += t1;
											//calc alignments
											CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
											CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));
											CUDA_SAFE_CALL(cudaMemcpy(bs_gpu,bs_cpu,mem_bs*sizeof(bool),cudaMemcpyHostToDevice));

											SW_CudaBS_Pair<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param,bs_gpu);

											//store results.
											//copy results
											post_proc_pair(helper1,pairs,t1,results_cpu,results_gpu,mem_result,reads,scaffold,parameters.corridor_with);

											//copy the left overs
											pre_filter_SW_pair_rest_BS(t1,liste,iter,helper1,reads_cpu,scaff_cpu,scaffold,parameters,bs_cpu);
										}
									}
								}
							}
						}
					}

				}
				scaffold.clear_scaff_keys();
			}
		}while(scaffold.sequence[0].size()!=0);
	}

	if(t1!=0) {
		amount+=t1;
		memset(&reads_cpu[t1*iter.seqlength],' ',(iter.proc-t1)*iter.seqlength*sizeof(char));
		//calc alignments
		CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
		CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));
		CUDA_SAFE_CALL(cudaMemcpy(bs_gpu,bs_cpu,mem_bs*sizeof(bool),cudaMemcpyHostToDevice));

		SW_CudaBS_Pair<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,matrix_gpu,results_gpu,param,bs_gpu);
		checkCUDAError("end2 1");

		post_proc_pair(helper1,pairs,t1, results_cpu,results_gpu,mem_result,reads,scaffold,parameters.corridor_with);
	}
	cout<<"\t\t\t"<<amount<<" where calculated"<<endl;


	BSlookupAG.clear();
	BSlookupTC.clear();
	delete [] parameters_cpu;
	delete [] results_cpu;
	delete [] bs_cpu;
	helper1.clear();

	delete [] reads_cpu;
	delete [] scaff_cpu;
	liste.clear();

	CUDA_SAFE_CALL(cudaFree(bs_gpu));
	CUDA_SAFE_CALL(cudaFree(matrix_gpu));
	CUDA_SAFE_CALL(cudaFree(results_gpu));
	CUDA_SAFE_CALL(cudaFree(reads_gpu));
	CUDA_SAFE_CALL(cudaFree(scaffold_gpu));
	CUDA_SAFE_CALL(cudaFree(param));

	checkCUDAError("after delete");
}
