/*
 * cuda_functions.cuh
 *
 *  Created on: Nov 26, 2009
 *      Author: fritz
 */

#ifndef CUDA_FUNCTIONS_CUH_
#define CUDA_FUNCTIONS_CUH_

#include "parser/parameter.h"

#  define CUDA_SAFE_CALL_NO_SYNC( call) do {                                \
    cudaError err = call;                                                    \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
        exit(EXIT_FAILURE);                                                  \
    } } while (0)
#  define CUDA_SAFE_CALL( call) do {                                        \
    CUDA_SAFE_CALL_NO_SYNC(call);                                            \
    cudaError err = cudaThreadSynchronize();                                 \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
        exit(EXIT_FAILURE);                                                  \
    } } while (0)
#  define CUT_SAFE_CALL( call)                                              \
    if( CUTTrue != call) {                                                   \
        fprintf(stderr, "Cut error in file '%s' in line %i.\n",              \
                __FILE__, __LINE__);                                         \
        exit(EXIT_FAILURE);                                                  \
    }

void checkCUDAError(const char *msg);

//void getSpecs(int & memsize, int & mps);
//
//void getDeviceProparties();
//
//void DisplayProperties(cudaDeviceProp * pDeviceProp);

void getDeviceProparties();
int cudaDevCount();
void initCudaParameter(Parameters & parameters);
void threadExit();

#endif /* CUDA_FUNCTIONS_CUH_ */
